"""Utilities."""
import functools
import inspect
import itertools
import re
from typing import (Any, Callable, Dict, List, Optional, Set, Tuple, Union,
                    get_type_hints)

import falcon

from falcontyping.typedjson import args_of, origin_of

from .exceptions import TypeValidationError

try:
    from typing_extensions import Protocol

    class ResourceMethodWithoutReturnValue(Protocol):

        def __call__(self, request: falcon.Request, respone: falcon.Response, **kwargs: Any) -> None:  # pragma: no cover
            ...

    class ResourceMethodWithReturnValue(Protocol):

        def __call__(self, request: falcon.Request, respone: falcon.Response, **kwargs: Any) -> Any:  # pragma: no cover
            ...

except ImportError:  # pragma: no cover
    ResourceMethodWithoutReturnValue = Callable  # type: ignore # pragma: no cover
    ResourceMethodWithReturnValue = Callable     # type: ignore # pragma: no cover

try:
    from pydantic import BaseModel as PydanticBaseModel

except ImportError:
    PydanticBaseModel = None  # type: ignore


try:
    from marshmallow import Schema as MarshmallowSchema

except ImportError:
    MarshmallowSchema = None  # type: ignore


_AVAILABLE_EXTERNAL_SCHEMA_PRODIVERS = set(
    filter(None, [MarshmallowSchema, PydanticBaseModel])
)

# Taken from https://github.com/falconry/falcon/blob/master/falcon/routing/compiled.py#L27
_FIELD_PATTERN = re.compile(
    # NOTE(kgriffs): This disallows the use of the '}' character within
    # an argstr. However, we don't really have a way of escaping
    # curly brackets in URI templates at the moment, so users should
    # see this as a similar restriction and so somewhat unsurprising.
    #
    # We may want to create a contextual parser at some point to
    # work around this problem.
    r'{((?P<fname>[^}:]*)((?P<cname_sep>:(?P<cname>[^}\(]*))(\((?P<argstr>[^}]*)\))?)?)}'
)

_METHOD_NAMES: List[str] = [
    'on_%s' % method for method in ('get', 'post', 'put', 'patch', 'delete', 'head', 'options')
]


def validate_type_preconditions(type_: Any) -> None:
    """Validate that used type hints are supported."""
    if len(_AVAILABLE_EXTERNAL_SCHEMA_PRODIVERS) == 0:
        raise TypeValidationError('No schema provider is installed. You need to install either Marshmallow or Pydantic')

    error = ('Resource methods must accept/return either Nothing, '
             'marshmallow.Schema or pydantic.BaseModel not {}')

    def predicate(type_: Any) -> bool:
        """Check if type can be used in the body of the request or response."""
        origin, args = origin_of(type_), args_of(type_)

        if origin is Union:
            return all(predicate(arg) for arg in args)

        else:
            has_supported_parent = any(
                issubclass(type_, parent) for parent in _AVAILABLE_EXTERNAL_SCHEMA_PRODIVERS
            )
            return (has_supported_parent or type_ == type(None)) and type_ is not Any  # noqa

    try:
        if type_ and not predicate(type_):
            raise TypeValidationError(error.format(type_))

    except TypeError:
        raise TypeValidationError(error.format(type_))


def validate_method_signature(method: ResourceMethodWithReturnValue, uri_parameters: Set[str]) -> Tuple[Optional[str],
                                                                                                        Dict]:
    """
    Validate whether resource method has the right signature.

    :returns: body parameter name and hints
    """
    hints = get_type_hints(method)
    arguments = inspect.getfullargspec(method).args

    for parameter in filter(lambda p: hints.get(p, None) is Any or p not in arguments, uri_parameters):
        raise TypeValidationError(f'URI parameter {parameter} has type hint Any or is missing from signature')

    validate_type_preconditions(
        hints.get('return', None)
    )

    if len(arguments) < 3:
        raise TypeValidationError('Every resource method must have the first two parameters as '
                                  'falcon.Request and falcon.Response')

    if hints.get(arguments[1], Any) not in [falcon.Request, Any]:
        raise TypeValidationError('First parameter must be of type falcon.Request')

    if hints.get(arguments[2], Any) not in [falcon.Response, Any]:
        raise TypeValidationError('Second parameter must be of type falcon.Response')

    body_parameters = set(arguments) - (uri_parameters | set(itertools.islice(arguments, 3)) | set(['return']))

    if len(body_parameters) > 1 and any(hints.get(parameter) for parameter in body_parameters):
        raise TypeValidationError('Any resource method can not accept more than one request parameter, that is one'
                                  'parameter of type marshmallow.Schema or pydantic.BaseModel')

    for parameter_name in body_parameters:
        validate_type_preconditions(hints.get(parameter_name, None))

    return next(iter(body_parameters & set(hints)), None), hints


def patch_resource_methods(uri_template: str, resource: Any) -> None:
    """Patch resource methods to add type hint supports."""
    uri_parameters = set(match.group('fname') for match in _FIELD_PATTERN.finditer(uri_template))

    def resource_method_wrapper(method: ResourceMethodWithReturnValue) -> ResourceMethodWithoutReturnValue:
        """
        Wraps resource methods.

        This function will takes whatever is returned by resource methods and assigns
        it to response.media.
        """
        @functools.wraps(method)
        def curried(request: falcon.Request, response: falcon.Response, **kwargs: Any) -> None:
            media = method(request, response, **kwargs)

            if media:
                response.media = media

        return curried

    resource.methods_body_parameter = {}
    resource.hints = {}

    for method_name in filter(functools.partial(hasattr, resource), _METHOD_NAMES):

        method: ResourceMethodWithReturnValue = getattr(resource, method_name)

        if not callable(method):
            raise TypeValidationError(f'{resource}.{method_name} must be a Callable')

        try:
            body_parameter, hints = validate_method_signature(method, uri_parameters=uri_parameters)

            resource.methods_body_parameter[method_name] = body_parameter
            resource.hints[method_name] = hints

        except TypeValidationError as e:
            raise TypeValidationError(f'{resource}.{method} raised: {e}')

        wrapped = resource_method_wrapper(method)

        setattr(resource, method_name, wrapped)



"""
The MIT License (MIT)

Copyright (c) 2018 Tomoya Kose

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This is a modified version of the annotation module at:
https://github.com/mitsuse/typedjson-python/blob/master/typedjson/annotation.py

Author: Tomoya Kose <tomoya@mitsuse.jp>
Repository: https://github.com/mitsuse/typedjson-python
"""
from typing import Dict, Optional, Tuple, Type


def args_of(type_: Type) -> Tuple[Type, ...]:
    args = getattr(type_, "__args__", None)
    return tuple() if args is None else args  # type: ignore


def hints_of(type_: Type) -> Optional[Dict[str, Type]]:
    from typing import get_type_hints
    from copy import copy

    origin = origin_of(type_)
    args = args_of(type_)
    type__ = type_ if origin is None else origin
    mapping = dict(zip(parameters_of(type_), args))

    if hasattr(type__, "__init__"):
        annotations = get_type_hints(type__.__init__)
        if len(mapping) > 0:
            annotations_: Dict[str, Type] = {}
            for n, t in annotations.items():
                if n == "return":
                    continue
                t_ = mapping.get(t)
                if t_ is None:
                    return None
                else:
                    annotations_[n] = t_
            return annotations_
        else:
            annotations_ = copy(annotations)
            annotations_.pop("return", None)
            return annotations_
    else:
        return None


def origin_of(type_: Type) -> Optional[Type]:
    from typing import List
    from typing import Tuple

    origin = getattr(type_, "__origin__", None)

    # In Python 3.6, the origin of Tuple type is `List` but in Python 3.7 it is `list`.
    if origin is List:
        return list
    # In Python 3.6, the origin of Tuple type is `Tuple` but in Python 3.7 it is `tuple`.
    elif origin is Tuple:
        return tuple
    else:
        return origin  # type: ignore


def parameters_of(type_: Type) -> Tuple[Type, ...]:
    origin = origin_of(type_)
    if origin is None:
        return tuple()
    else:
        parameters = getattr(origin, "__parameters__", None)
        return tuple() if parameters is None else parameters  # type: ignore



"""API wrapper."""
from typing import Any

import falcon
import wrapt

from ..middleware import TypingMiddleware
from .resource import TypedResource
from .utils import patch_resource_methods


class TypedAPI(wrapt.CallableObjectProxy):
    """Transparent proxy over falcon.API."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        middleware = kwargs.get('middleware', [])

        if not any(isinstance(m, TypingMiddleware) for m in middleware):
            kwargs = {**kwargs, 'middleware': middleware + [TypingMiddleware()]}

        super().__init__(
            falcon.API(*args, **kwargs)
        )

    def add_route(self, uri_template: str, resource: Any, **kwargs: Any) -> Any:
        """Extract type hints information before adding route."""
        if isinstance(resource, TypedResource):
            patch_resource_methods(uri_template, resource)

        return self.__wrapped__.add_route(uri_template, resource, **kwargs)



"""Typing middleware."""
from typing import Any, Callable, Dict, Optional

import falcon
from falcon import Request, Response

from falcontyping.base import (PydanticBaseModel, TypedResource,
                               TypeValidationError)
from falcontyping.typedjson import DecodingError, ExternalSerializerException
from falcontyping.typedjson import decode as decode_using_hints

_VALID_RESPONSE_TYPES = set([PydanticBaseModel, dict, type(None)])


class TypingMiddleware:  # pragma: no cover

    @staticmethod
    def _decode_or_raise_error(hint: Any, parameter: Any) -> Any:
        """
        Decode value using type hint or fail.

        :raises: falcon.HTTPError or ExternalSerializerException
        """
        result = decode_using_hints(hint, parameter)

        if isinstance(result, DecodingError):
            if isinstance(result.reason, ExternalSerializerException):
                raise result.reason.exception from None

            else:
                raise falcon.HTTPError(status=falcon.HTTP_UNPROCESSABLE_ENTITY,  # pylint: disable=no-member
                                       description=f'\'{parameter}\' must be of type {hint} not {type(parameter)}')

        return result

    @staticmethod
    def _try_decode_query_or_body(request: falcon.Request, hint: Any) -> Any:
        """Decode values by looking for them in both URI and request body."""
        # An assumption is being made here, That only POST, PUT and PATCH can have bodies.
        if request.method.lower() in ['post', 'put', 'patch']:
            key = 'media'

        else:
            key = 'params'

        return TypingMiddleware._decode_or_raise_error(hint, getattr(request, key, None))

    def process_request(self, request: Request, response: Response) -> None:
        """
        Process the request before routing it.

        Because Falcon routes each request based on req.path, a
        request can be effectively re-routed by setting that
        attribute to a new value from within process_request().

        :param request: Request object that will eventually be
            routed to an on_* responder method.
        :param response: Response object that will be routed to
            the on_* responder.
        """
        ...

    def process_resource(self, request: Request, response: Response, resource: Any, parameters: Dict) -> None:
        """
        Process the request after routing.

        This method is only called when the request matches
        a route to a resource.

        :param request: Request object that will be passed to the
            routed responder.
        :param response: Response object that will be passed to the
            responder.
        :param resource: Resource object to which the request was
            routed.
        :param parameters: A dict-like object representing any additional
            parameters derived from the route's URI template fields,
            that will be passed to the resource's responder
            method as keyword arguments.
        """
        if not isinstance(resource, TypedResource):
            return

        handler: Optional[Callable] = getattr(resource, 'on_%s' % request.method.lower(), None)

        if handler:
            # Get hints for only those variables that should be passed to the request handler.
            hints = resource.hints[handler.__name__]

            # Decode values using type hints, All values in parameters will be based as
            # Keyword arguments to the request handler.
            for parameter in filter(hints.get, parameters):
                parameters[parameter] = self._decode_or_raise_error(hints[parameter], parameters.get(parameter))

            # Decode body parameter if there is one.
            body_parameter = resource.methods_body_parameter[handler.__name__]
            if body_parameter:
                parameters[body_parameter] = self._try_decode_query_or_body(request, hints[body_parameter])

    def process_response(self, request: Request, response: Response, resource: Any, request_succeeded: bool) -> None:
        """
        Post-processing of the response (after routing).

        :param request: Request object.
        :param response: Response object.
        :param resource: Resource object to which the request was routed.
            May be None if no route was found for the request.

        :param request_succeeded: True if no exceptions were raised while the framework processed and
            routed the request; otherwise False.
        """
        if not (isinstance(resource, TypedResource) and request_succeeded):
            return

        handler: Optional[Callable] = getattr(resource, 'on_%s' % request.method.lower(), None)
        # Get type hint for the return type of the request handler.
        hint: Any = resource.hints[handler.__name__].get('return') if handler else None

        if hint:
            media = getattr(response, 'media', None)
            media = decode_using_hints(hint, media)

            if not any(isinstance(media, type_) for type_ in _VALID_RESPONSE_TYPES):  # type: ignore
                raise TypeValidationError(f'{resource}.{handler} returned a unexpected value. ',
                                          f'Resource methods must return either Nothing, '
                                          f'marshmallow.Schema or pydantic.BaseModel not {type(media)}')

            if isinstance(media, PydanticBaseModel):
                media = media.dict()

            response.media = media



"""Setup module."""
from setuptools import setup

with open('README.md', 'r', encoding='UTF-8') as file:
    README = file.read()

setup(
    name='falcontyping',
    version='0.7.0',
    url='https://github.com/abdelrahman-t/falcontyping',
    author='Abdurrahman Talaat <abdurrahman.talaat@gmail.com>',
    author_email='abdurrahman.talaat@gmail.com',
    packages=['falcontyping', 'falcontyping.base', 'falcontyping.middleware', 'falcontyping.typedjson'],
    include_package_data=True,
    python_requires='>=3.7.0',
    install_requires=['falcon', 'wrapt'],

    description='Add type hints support to Falcon with Pydantic and Marshmallow integration',
    long_description=README,
    long_description_content_type='text/markdown',

    keywords=['falcon', 'typing', 'hints', 'mypy', 'pydantic', 'marshmallow'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.7',
    ],

    project_urls={
        'Documentation': 'https://github.com/abdelrahman-t/falcontyping'
    },
)



from .exceptions import TypeValidationError
from .resource import TypedResource
from .utils import _AVAILABLE_EXTERNAL_SCHEMA_PRODIVERS as schema_providers
from .utils import MarshmallowSchema, PydanticBaseModel
from .wrapper import TypedAPI

__all__ = ['TypedResource', 'TypeValidationError', 'TypedAPI',
           'schema_providers', 'PydanticBaseModel', 'MarshmallowSchema']



"""
The MIT License (MIT)

Copyright (c) 2018 Tomoya Kose

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This is a modified version of the decoding module at:
https://github.com/mitsuse/typedjson-python/blob/master/typedjson/decoding.py

Author: Tomoya Kose <tomoya@mitsuse.jp>
Repository: https://github.com/mitsuse/typedjson-python
"""
from typing import Any, Type, TypeVar, Union

from .base import (Decoded, DecodingError, ExternalSerializerException, Path,
                   TypeMismatch, UnsupportedDecoding)

try:
    from pydantic import BaseModel as PydanticBaseModel

except ImportError:
    PydanticBaseModel = None  # type: ignore


try:
    from marshmallow import Schema as MarshmallowSchema

except ImportError:
    MarshmallowSchema = None  # type: ignore

_PRIMITIVE_TYPES = (str, float, int, bool, type(None))


def parse_int_from_string(json: str, path: Path) -> Union[Decoded, DecodingError]:
    try:
        number = float(json)

        if not number.is_integer():
            raise ValueError

        return int(number)  # type: ignore

    except ValueError:
        return DecodingError(TypeMismatch(path))


def parse_pydantic_from_dict(type_: Any, json: Any, path: Path) -> Union[Decoded, DecodingError]:
    try:
        return type_(**json)  # type: ignore

    except Exception as e:
        return DecodingError(reason=ExternalSerializerException(path, exception=e))


def parse_marshmallow_from_dict(type_: Any, json: Any, path: Path) -> Union[Decoded, DecodingError]:
    try:
        return type_().load(json)  # type: ignore

    except Exception as e:
        return DecodingError(reason=ExternalSerializerException(path, exception=e))


def decode(type_: Type[Decoded], json: Any, path: Path = ()) -> Union[Decoded, DecodingError]:
    decoders = (
        decode_as_union,
        decode_as_primitive,
        decode_as_model
    )

    result_final: Union[Decoded, DecodingError] = DecodingError(
        UnsupportedDecoding(path)
    )
    for decoder in decoders:
        result = decoder(type_, json, path)

        if isinstance(result, DecodingError):
            if isinstance(result.reason, TypeMismatch):
                result_final = result
        else:
            result_final = result
            break

    return result_final


def decode_as_model(type_: Type[Decoded], json: Any, path: Path) -> Union[Decoded, DecodingError]:
    from .annotation import origin_of

    if origin_of(type_) in (Union, Any, ):
        return DecodingError(UnsupportedDecoding(path))

    if isinstance(json, type_):
        return json

    if PydanticBaseModel and issubclass(type_, PydanticBaseModel):
        if json:
            return parse_pydantic_from_dict(type_, json=json, path=path)

        else:
            return DecodingError(TypeMismatch(path))

    elif MarshmallowSchema and issubclass(type_, MarshmallowSchema):
        if json:
            return parse_marshmallow_from_dict(type_, json=json, path=path)

        else:
            return DecodingError(TypeMismatch(path))

    else:
        return DecodingError(UnsupportedDecoding(path))


def decode_as_primitive(type_: Type[Decoded], json: Any, path: Path) -> Union[Decoded, DecodingError]:
    if type_ not in _PRIMITIVE_TYPES:
        return DecodingError(UnsupportedDecoding(path))

    if isinstance(json, type_):
        return json

    if type_ is int and isinstance(json, str):
        return parse_int_from_string(json, path)

    else:
        return DecodingError(TypeMismatch(path))


def decode_as_union(type_: Type[Decoded], json: Any, path: Path) -> Union[Decoded, DecodingError]:
    from .annotation import args_of
    from .annotation import origin_of

    if origin_of(type_) is Union:
        args = args_of(type_)

        for type_ in args:
            if type_.__class__ is TypeVar:
                return DecodingError(UnsupportedDecoding(path))

        for type_ in args:
            decoded = decode(type_, json, path)

            if not isinstance(decoded, DecodingError):
                break

        return decoded
    else:
        return DecodingError(UnsupportedDecoding(path))
